/*
 * Copyright (c) 2019 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdio.h>
#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/logging/log.h>
#include <modem/nrf_modem_lib.h>
#include <nrf_modem_at.h>
#include <zephyr/net/socket.h>
#include "CoAP/CoAP.h"

LOG_MODULE_REGISTER(main, LOG_LEVEL_INF);

char response[512];
uint8_t RecvBuff[512];
// TCP Server Implementation ----------

static int SeverRcevHandler(int client)
{
	int rc;
    int errno;

	struct pollfd fd = {
		.fd = client,
		.events = POLLIN
	};

	rc = poll(&fd, 1, MSEC_PER_SEC * 10);
	if (rc < 0) {
		LOG_WRN("poll() error: %d", -errno);
		return -errno;
	} else if (rc == 0) {
		LOG_WRN("poll() timeout");
		return -EAGAIN;
	}

	rc = recv(client, RecvBuff, sizeof(RecvBuff), 0);
	if (rc <= 0) {
		return rc == 0 ? -ENOTCONN : -errno;
	}

	LOG_HEXDUMP_DBG(RecvBuff, sizeof(RecvBuff), "RecvBuff:>");


    if (send(client, RecvBuff, strlen(RecvBuff), 0) < 0) {
		return -errno;
	}

	return errno;
}

int StartTCPServer(void)
{
	int serv;
	struct sockaddr_in bind_addr;
	static int counter;
	int err = 0;
    char *Server_Addr_str = "70.25.147.111";
    uint32_t Server_Addr_u32 = 0;
    uint16_t Server_port = 502;
    char *next = 0;

	LOG_INF("[%d]%s", __LINE__, __func__);

	serv = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (serv < 0) {
		LOG_ERR("error: socket: %d", errno);
		return 0;
	}

	LOG_INF("[%d]%s >> socket >> serv >> %d", __LINE__, __func__, serv);

    Server_Addr_u32 = strtol(Server_Addr_str, &next, 10);
    Server_Addr_u32 = Server_Addr_u32 << 8;
    Server_Addr_u32 |= strtol(&next[1], &next, 10);
    Server_Addr_u32 = Server_Addr_u32 << 8;
    Server_Addr_u32 |= strtol(&next[1], &next, 10);
    Server_Addr_u32 = Server_Addr_u32 << 8;
    Server_Addr_u32 |= strtol(&next[1], &next, 10);

	bind_addr.sin_family = AF_INET;
	bind_addr.sin_addr.s_addr = htonl(Server_Addr_u32);// htonl(INADDR_ANY); // INADDR_ANY
	bind_addr.sin_port = htons(Server_port);
    
	if (bind(serv, (struct sockaddr *)&bind_addr, sizeof(bind_addr)) < 0) {
		LOG_ERR("error: bind: %d", errno);
		return 0;
	}

	if (listen(serv, 5) < 0) {
		LOG_ERR("error: listen: %d", errno);
		return 0;
	}

	errno = 0;
	LOG_INF("Started TCP server example on port %d, %d", Server_port, errno);
	LOG_DBG("Entering infinite while Loop %d", __LINE__);

	while (1) {
		struct sockaddr_in client_addr;
		socklen_t client_addr_len = sizeof(client_addr);
		char addr_str[INET_ADDRSTRLEN];
		int client;
		int rc;

		client = accept(serv, (struct sockaddr *)&client_addr,
				&client_addr_len);

		if (client < 0) {
			LOG_ERR("error: accept: %d, %d", errno, client);
			k_msleep(200);
			continue;
		}

		inet_ntop(client_addr.sin_family, &client_addr.sin_addr,
			  addr_str, sizeof(addr_str));
		LOG_INF("Connection #%d from %s",
			counter++, addr_str);

		do {
			rc = SeverRcevHandler(client);
			k_msleep(1);
		} while (!rc);

		close(client);
		LOG_INF("Connection from %s closed, errno %d",
			addr_str, rc);
	}
	return 0;
}


void main(void)
{
    int err = 0;

    LOG_INF("[%d]%s", __LINE__, __func__);
	LOG_INF("Project : CoAPAndServer");

    err = nrf_modem_lib_init();
    if (err < 0) 
    {
        LOG_ERR("Modem library init failed, err: %d", err);
        return;
    }

    k_msleep(1000);

    err = nrf_modem_at_cmd(response, sizeof(response), "AT+CFUN=1");
    if (err) // This condition will true, When there is an error occured
    {
        LOG_ERR("CFUN Failed %d", err);
        return;
    }

    do  // Check Network in do..while();
    {
        k_sleep(K_SECONDS(1)); // Delay between two iteration
        err = nrf_modem_at_cmd(response, sizeof(response), "AT+CEREG?");
        if (err) // true, when error
        {
            LOG_ERR("ERR : %d, %s", err, response);
        }
        
        LOG_INF("RSP : %s", response);
        
    } while(!strstr(response, "+CEREG: 0,1"));

    CoAPSetup();

    k_msleep(1000);

    StartTCPServer();

    while(1)
    { 
        // RTC_Read(); // RTC Read test
        // SHT40_Test(); // SHT40, NOTE : check mutex lock
        // vI2cScan(); // NOTE : check mutex lock
        
        
        // char DisplayValue[20] = {0};

        // snprintf(DisplayValue, 20, "%.1f C\n%.1f RH\n%.1f H", fReadTemprature(), fReadHumidity(), fReadHumidex());
        // DisplayString(DisplayValue);
        // LOG_INF("%s", DisplayValue);

        k_msleep(5000);
    }
}


/* [ Thread Tree ]
    Start_SysInit_Thread() --> Start_Network_Check_Thread() // Create new thread and distory self
    Start_Network_Check_Thread() --> Start_Get_APN_Thread((uint32_t)1000); // Create new thread and distory self
    Start_Get_APN_Thread((uint32_t)1000); --> Start_ModbusServer() // Create new thread and distory self
    Start_ModbusServer() // This one runs forever
*/